/* kexec-loader - Keymap loading code
 * Copyright (C) 2007-2009 Daniel Collins <solemnwarning@solemnwarning.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include <sys/ioctl.h>

#include "console.h"
#include "misc.h"

struct action_code {
	char const *name;
	unsigned short code;
};

static struct action_code codes[] = {
	{"nul", 0x0000},
	{"Control_a", 0x0001},
	{"Control_b", 0x0002},
	{"Control_c", 0x0003},
	{"Control_d", 0x0004},
	{"Control_e", 0x0005},
	{"Control_f", 0x0006},
	{"Control_g", 0x0007},
	{"BackSpace", 0x0008},
	{"Tab", 0x0009},
	{"Linefeed", 0x000a},
	{"Control_k", 0x000b},
	{"Control_l", 0x000c},
	{"Control_m", 0x000d},
	{"Control_n", 0x000e},
	{"Control_o", 0x000f},
	{"Control_p", 0x0010},
	{"Control_q", 0x0011},
	{"Control_r", 0x0012},
	{"Control_s", 0x0013},
	{"Control_t", 0x0014},
	{"Control_u", 0x0015},
	{"Control_v", 0x0016},
	{"Control_w", 0x0017},
	{"Control_x", 0x0018},
	{"Control_y", 0x0019},
	{"Control_z", 0x001a},
	{"Escape", 0x001b},
	{"Control_backslash", 0x001c},
	{"Control_bracketright", 0x001d},
	{"Control_asciicircum", 0x001e},
	{"Control_underscore", 0x001f},
	{"space", 0x0020},
	{"exclam", 0x0021},
	{"quotedbl", 0x0022},
	{"numbersign", 0x0023},
	{"dollar", 0x0024},
	{"percent", 0x0025},
	{"ampersand", 0x0026},
	{"apostrophe", 0x0027},
	{"parenleft", 0x0028},
	{"parenright", 0x0029},
	{"asterisk", 0x002a},
	{"plus", 0x002b},
	{"comma", 0x002c},
	{"minus", 0x002d},
	{"period", 0x002e},
	{"slash", 0x002f},
	{"zero", 0x0030},
	{"one", 0x0031},
	{"two", 0x0032},
	{"three", 0x0033},
	{"four", 0x0034},
	{"five", 0x0035},
	{"six", 0x0036},
	{"seven", 0x0037},
	{"eight", 0x0038},
	{"nine", 0x0039},
	{"colon", 0x003a},
	{"semicolon", 0x003b},
	{"less", 0x003c},
	{"equal", 0x003d},
	{"greater", 0x003e},
	{"question", 0x003f},
	{"at", 0x0040},
	{"A", 0x0041},
	{"B", 0x0042},
	{"C", 0x0043},
	{"D", 0x0044},
	{"E", 0x0045},
	{"F", 0x0046},
	{"G", 0x0047},
	{"H", 0x0048},
	{"I", 0x0049},
	{"J", 0x004a},
	{"K", 0x004b},
	{"L", 0x004c},
	{"M", 0x004d},
	{"N", 0x004e},
	{"O", 0x004f},
	{"P", 0x0050},
	{"Q", 0x0051},
	{"R", 0x0052},
	{"S", 0x0053},
	{"T", 0x0054},
	{"U", 0x0055},
	{"V", 0x0056},
	{"W", 0x0057},
	{"X", 0x0058},
	{"Y", 0x0059},
	{"Z", 0x005a},
	{"bracketleft", 0x005b},
	{"backslash", 0x005c},
	{"bracketright", 0x005d},
	{"asciicircum", 0x005e},
	{"underscore", 0x005f},
	{"grave", 0x0060},
	{"a", 0x0061},
	{"b", 0x0062},
	{"c", 0x0063},
	{"d", 0x0064},
	{"e", 0x0065},
	{"f", 0x0066},
	{"g", 0x0067},
	{"h", 0x0068},
	{"i", 0x0069},
	{"j", 0x006a},
	{"k", 0x006b},
	{"l", 0x006c},
	{"m", 0x006d},
	{"n", 0x006e},
	{"o", 0x006f},
	{"p", 0x0070},
	{"q", 0x0071},
	{"r", 0x0072},
	{"s", 0x0073},
	{"t", 0x0074},
	{"u", 0x0075},
	{"v", 0x0076},
	{"w", 0x0077},
	{"x", 0x0078},
	{"y", 0x0079},
	{"z", 0x007a},
	{"braceleft", 0x007b},
	{"bar", 0x007c},
	{"braceright", 0x007d},
	{"asciitilde", 0x007e},
	{"Delete", 0x007f},
	{"nobreakspace", 0x00a0},
	{"exclamdown", 0x00a1},
	{"cent", 0x00a2},
	{"sterling", 0x00a3},
	{"currency", 0x00a4},
	{"yen", 0x00a5},
	{"brokenbar", 0x00a6},
	{"section", 0x00a7},
	{"diaeresis", 0x00a8},
	{"copyright", 0x00a9},
	{"ordfeminine", 0x00aa},
	{"guillemotleft", 0x00ab},
	{"notsign", 0x00ac},
	{"hyphen", 0x00ad},
	{"registered", 0x00ae},
	{"macron", 0x00af},
	{"degree", 0x00b0},
	{"plusminus", 0x00b1},
	{"twosuperior", 0x00b2},
	{"threesuperior", 0x00b3},
	{"acute", 0x00b4},
	{"mu", 0x00b5},
	{"paragraph", 0x00b6},
	{"periodcentered", 0x00b7},
	{"cedilla", 0x00b8},
	{"onesuperior", 0x00b9},
	{"masculine", 0x00ba},
	{"guillemotright", 0x00bb},
	{"onequarter", 0x00bc},
	{"onehalf", 0x00bd},
	{"threequarters", 0x00be},
	{"questiondown", 0x00bf},
	{"Agrave", 0x00c0},
	{"Aacute", 0x00c1},
	{"Acircumflex", 0x00c2},
	{"Atilde", 0x00c3},
	{"Adiaeresis", 0x00c4},
	{"Aring", 0x00c5},
	{"AE", 0x00c6},
	{"Ccedilla", 0x00c7},
	{"Egrave", 0x00c8},
	{"Eacute", 0x00c9},
	{"Ecircumflex", 0x00ca},
	{"Ediaeresis", 0x00cb},
	{"Igrave", 0x00cc},
	{"Iacute", 0x00cd},
	{"Icircumflex", 0x00ce},
	{"Idiaeresis", 0x00cf},
	{"ETH", 0x00d0},
	{"Ntilde", 0x00d1},
	{"Ograve", 0x00d2},
	{"Oacute", 0x00d3},
	{"Ocircumflex", 0x00d4},
	{"Otilde", 0x00d5},
	{"Odiaeresis", 0x00d6},
	{"multiply", 0x00d7},
	{"Ooblique", 0x00d8},
	{"Ugrave", 0x00d9},
	{"Uacute", 0x00da},
	{"Ucircumflex", 0x00db},
	{"Udiaeresis", 0x00dc},
	{"Yacute", 0x00dd},
	{"THORN", 0x00de},
	{"ssharp", 0x00df},
	{"agrave", 0x00e0},
	{"aacute", 0x00e1},
	{"acircumflex", 0x00e2},
	{"atilde", 0x00e3},
	{"adiaeresis", 0x00e4},
	{"aring", 0x00e5},
	{"ae", 0x00e6},
	{"ccedilla", 0x00e7},
	{"egrave", 0x00e8},
	{"eacute", 0x00e9},
	{"ecircumflex", 0x00ea},
	{"ediaeresis", 0x00eb},
	{"igrave", 0x00ec},
	{"iacute", 0x00ed},
	{"icircumflex", 0x00ee},
	{"idiaeresis", 0x00ef},
	{"eth", 0x00f0},
	{"ntilde", 0x00f1},
	{"ograve", 0x00f2},
	{"oacute", 0x00f3},
	{"ocircumflex", 0x00f4},
	{"otilde", 0x00f5},
	{"odiaeresis", 0x00f6},
	{"division", 0x00f7},
	{"oslash", 0x00f8},
	{"ugrave", 0x00f9},
	{"uacute", 0x00fa},
	{"ucircumflex", 0x00fb},
	{"udiaeresis", 0x00fc},
	{"yacute", 0x00fd},
	{"thorn", 0x00fe},
	{"ydiaeresis", 0x00ff},
	{"F1", 0x0100},
	{"F2", 0x0101},
	{"F3", 0x0102},
	{"F4", 0x0103},
	{"F5", 0x0104},
	{"F6", 0x0105},
	{"F7", 0x0106},
	{"F8", 0x0107},
	{"F9", 0x0108},
	{"F10", 0x0109},
	{"F11", 0x010a},
	{"F12", 0x010b},
	{"F13", 0x010c},
	{"F14", 0x010d},
	{"F15", 0x010e},
	{"F16", 0x010f},
	{"F17", 0x0110},
	{"F18", 0x0111},
	{"F19", 0x0112},
	{"F20", 0x0113},
	{"Find", 0x0114},
	{"Insert", 0x0115},
	{"Remove", 0x0116},
	{"Select", 0x0117},
	{"Prior", 0x0118},
	{"Next", 0x0119},
	{"Macro", 0x011a},
	{"Help", 0x011b},
	{"Do", 0x011c},
	{"Pause", 0x011d},
	{"F21", 0x011e},
	{"F22", 0x011f},
	{"F23", 0x0120},
	{"F24", 0x0121},
	{"F25", 0x0122},
	{"F26", 0x0123},
	{"F27", 0x0124},
	{"F28", 0x0125},
	{"F29", 0x0126},
	{"F30", 0x0127},
	{"F31", 0x0128},
	{"F32", 0x0129},
	{"F33", 0x012a},
	{"F34", 0x012b},
	{"F35", 0x012c},
	{"F36", 0x012d},
	{"VoidSymbol", 0x0200},
	{"Return", 0x0201},
	{"Show_Registers", 0x0202},
	{"Show_Memory", 0x0203},
	{"Show_State", 0x0204},
	{"Break", 0x0205},
	{"Last_Console", 0x0206},
	{"Caps_Lock", 0x0207},
	{"Num_Lock", 0x0208},
	{"Scroll_Lock", 0x0209},
	{"Scroll_Forward", 0x020a},
	{"Scroll_Backward", 0x020b},
	{"Boot", 0x020c},
	{"Caps_On", 0x020d},
	{"Compose", 0x020e},
	{"SAK", 0x020f},
	{"Decr_Console", 0x0210},
	{"Incr_Console", 0x0211},
	{"KeyboardSignal", 0x0212},
	{"Bare_Num_Lock", 0x0213},
	{"KP_0", 0x0300},
	{"KP_1", 0x0301},
	{"KP_2", 0x0302},
	{"KP_3", 0x0303},
	{"KP_4", 0x0304},
	{"KP_5", 0x0305},
	{"KP_6", 0x0306},
	{"KP_7", 0x0307},
	{"KP_8", 0x0308},
	{"KP_9", 0x0309},
	{"KP_Add", 0x030a},
	{"KP_Subtract", 0x030b},
	{"KP_Multiply", 0x030c},
	{"KP_Divide", 0x030d},
	{"KP_Enter", 0x030e},
	{"KP_Comma", 0x030f},
	{"KP_Period", 0x0310},
	{"KP_MinPlus", 0x0311},
	{"dead_grave", 0x0400},
	{"dead_acute", 0x0401},
	{"dead_circumflex", 0x0402},
	{"dead_tilde", 0x0403},
	{"dead_diaeresis", 0x0404},
	{"dead_cedilla", 0x0405},
	{"Console_1", 0x0500},
	{"Console_2", 0x0501},
	{"Console_3", 0x0502},
	{"Console_4", 0x0503},
	{"Console_5", 0x0504},
	{"Console_6", 0x0505},
	{"Console_7", 0x0506},
	{"Console_8", 0x0507},
	{"Console_9", 0x0508},
	{"Console_10", 0x0509},
	{"Console_11", 0x050a},
	{"Console_12", 0x050b},
	{"Console_13", 0x050c},
	{"Console_14", 0x050d},
	{"Console_15", 0x050e},
	{"Console_16", 0x050f},
	{"Console_17", 0x0510},
	{"Console_18", 0x0511},
	{"Console_19", 0x0512},
	{"Console_20", 0x0513},
	{"Console_21", 0x0514},
	{"Console_22", 0x0515},
	{"Console_23", 0x0516},
	{"Console_24", 0x0517},
	{"Console_25", 0x0518},
	{"Console_26", 0x0519},
	{"Console_27", 0x051a},
	{"Console_28", 0x051b},
	{"Console_29", 0x051c},
	{"Console_30", 0x051d},
	{"Console_31", 0x051e},
	{"Console_32", 0x051f},
	{"Console_33", 0x0520},
	{"Console_34", 0x0521},
	{"Console_35", 0x0522},
	{"Console_36", 0x0523},
	{"Console_37", 0x0524},
	{"Console_38", 0x0525},
	{"Console_39", 0x0526},
	{"Console_40", 0x0527},
	{"Console_41", 0x0528},
	{"Console_42", 0x0529},
	{"Console_43", 0x052a},
	{"Console_44", 0x052b},
	{"Console_45", 0x052c},
	{"Console_46", 0x052d},
	{"Console_47", 0x052e},
	{"Console_48", 0x052f},
	{"Console_49", 0x0530},
	{"Console_50", 0x0531},
	{"Console_51", 0x0532},
	{"Console_52", 0x0533},
	{"Console_53", 0x0534},
	{"Console_54", 0x0535},
	{"Console_55", 0x0536},
	{"Console_56", 0x0537},
	{"Console_57", 0x0538},
	{"Console_58", 0x0539},
	{"Console_59", 0x053a},
	{"Console_60", 0x053b},
	{"Console_61", 0x053c},
	{"Console_62", 0x053d},
	{"Console_63", 0x053e},
	{"Down", 0x0600},
	{"Left", 0x0601},
	{"Right", 0x0602},
	{"Up", 0x0603},
	{"Shift", 0x0700},
	{"AltGr", 0x0701},
	{"Control", 0x0702},
	{"Alt", 0x0703},
	{"ShiftL", 0x0704},
	{"ShiftR", 0x0705},
	{"CtrlL", 0x0706},
	{"CtrlR", 0x0707},
	{"CapsShift", 0x0708},
	{"Meta_nul", 0x0800},
	{"Meta_Control_a", 0x0801},
	{"Meta_Control_b", 0x0802},
	{"Meta_Control_c", 0x0803},
	{"Meta_Control_d", 0x0804},
	{"Meta_Control_e", 0x0805},
	{"Meta_Control_f", 0x0806},
	{"Meta_Control_g", 0x0807},
	{"Meta_BackSpace", 0x0808},
	{"Meta_Tab", 0x0809},
	{"Meta_Linefeed", 0x080a},
	{"Meta_Control_k", 0x080b},
	{"Meta_Control_l", 0x080c},
	{"Meta_Control_m", 0x080d},
	{"Meta_Control_n", 0x080e},
	{"Meta_Control_o", 0x080f},
	{"Meta_Control_p", 0x0810},
	{"Meta_Control_q", 0x0811},
	{"Meta_Control_r", 0x0812},
	{"Meta_Control_s", 0x0813},
	{"Meta_Control_t", 0x0814},
	{"Meta_Control_u", 0x0815},
	{"Meta_Control_v", 0x0816},
	{"Meta_Control_w", 0x0817},
	{"Meta_Control_x", 0x0818},
	{"Meta_Control_y", 0x0819},
	{"Meta_Control_z", 0x081a},
	{"Meta_Escape", 0x081b},
	{"Meta_Control_backslash", 0x081c},
	{"Meta_Control_bracketright", 0x081d},
	{"Meta_Control_asciicircum", 0x081e},
	{"Meta_Control_underscore", 0x081f},
	{"Meta_space", 0x0820},
	{"Meta_exclam", 0x0821},
	{"Meta_quotedbl", 0x0822},
	{"Meta_numbersign", 0x0823},
	{"Meta_dollar", 0x0824},
	{"Meta_percent", 0x0825},
	{"Meta_ampersand", 0x0826},
	{"Meta_apostrophe", 0x0827},
	{"Meta_parenleft", 0x0828},
	{"Meta_parenright", 0x0829},
	{"Meta_asterisk", 0x082a},
	{"Meta_plus", 0x082b},
	{"Meta_comma", 0x082c},
	{"Meta_minus", 0x082d},
	{"Meta_period", 0x082e},
	{"Meta_slash", 0x082f},
	{"Meta_zero", 0x0830},
	{"Meta_one", 0x0831},
	{"Meta_two", 0x0832},
	{"Meta_three", 0x0833},
	{"Meta_four", 0x0834},
	{"Meta_five", 0x0835},
	{"Meta_six", 0x0836},
	{"Meta_seven", 0x0837},
	{"Meta_eight", 0x0838},
	{"Meta_nine", 0x0839},
	{"Meta_colon", 0x083a},
	{"Meta_semicolon", 0x083b},
	{"Meta_less", 0x083c},
	{"Meta_equal", 0x083d},
	{"Meta_greater", 0x083e},
	{"Meta_question", 0x083f},
	{"Meta_at", 0x0840},
	{"Meta_A", 0x0841},
	{"Meta_B", 0x0842},
	{"Meta_C", 0x0843},
	{"Meta_D", 0x0844},
	{"Meta_E", 0x0845},
	{"Meta_F", 0x0846},
	{"Meta_G", 0x0847},
	{"Meta_H", 0x0848},
	{"Meta_I", 0x0849},
	{"Meta_J", 0x084a},
	{"Meta_K", 0x084b},
	{"Meta_L", 0x084c},
	{"Meta_M", 0x084d},
	{"Meta_N", 0x084e},
	{"Meta_O", 0x084f},
	{"Meta_P", 0x0850},
	{"Meta_Q", 0x0851},
	{"Meta_R", 0x0852},
	{"Meta_S", 0x0853},
	{"Meta_T", 0x0854},
	{"Meta_U", 0x0855},
	{"Meta_V", 0x0856},
	{"Meta_W", 0x0857},
	{"Meta_X", 0x0858},
	{"Meta_Y", 0x0859},
	{"Meta_Z", 0x085a},
	{"Meta_bracketleft", 0x085b},
	{"Meta_backslash", 0x085c},
	{"Meta_bracketright", 0x085d},
	{"Meta_asciicircum", 0x085e},
	{"Meta_underscore", 0x085f},
	{"Meta_grave", 0x0860},
	{"Meta_a", 0x0861},
	{"Meta_b", 0x0862},
	{"Meta_c", 0x0863},
	{"Meta_d", 0x0864},
	{"Meta_e", 0x0865},
	{"Meta_f", 0x0866},
	{"Meta_g", 0x0867},
	{"Meta_h", 0x0868},
	{"Meta_i", 0x0869},
	{"Meta_j", 0x086a},
	{"Meta_k", 0x086b},
	{"Meta_l", 0x086c},
	{"Meta_m", 0x086d},
	{"Meta_n", 0x086e},
	{"Meta_o", 0x086f},
	{"Meta_p", 0x0870},
	{"Meta_q", 0x0871},
	{"Meta_r", 0x0872},
	{"Meta_s", 0x0873},
	{"Meta_t", 0x0874},
	{"Meta_u", 0x0875},
	{"Meta_v", 0x0876},
	{"Meta_w", 0x0877},
	{"Meta_x", 0x0878},
	{"Meta_y", 0x0879},
	{"Meta_z", 0x087a},
	{"Meta_braceleft", 0x087b},
	{"Meta_bar", 0x087c},
	{"Meta_braceright", 0x087d},
	{"Meta_asciitilde", 0x087e},
	{"Meta_Delete", 0x087f},
	{"Ascii_0", 0x0900},
	{"Ascii_1", 0x0901},
	{"Ascii_2", 0x0902},
	{"Ascii_3", 0x0903},
	{"Ascii_4", 0x0904},
	{"Ascii_5", 0x0905},
	{"Ascii_6", 0x0906},
	{"Ascii_7", 0x0907},
	{"Ascii_8", 0x0908},
	{"Ascii_9", 0x0909},
	{"Hex_0", 0x090a},
	{"Hex_1", 0x090b},
	{"Hex_2", 0x090c},
	{"Hex_3", 0x090d},
	{"Hex_4", 0x090e},
	{"Hex_5", 0x090f},
	{"Hex_6", 0x0910},
	{"Hex_7", 0x0911},
	{"Hex_8", 0x0912},
	{"Hex_9", 0x0913},
	{"Hex_A", 0x0914},
	{"Hex_B", 0x0915},
	{"Hex_C", 0x0916},
	{"Hex_D", 0x0917},
	{"Hex_E", 0x0918},
	{"Hex_F", 0x0919},
	{"Shift_Lock", 0x0a00},
	{"AltGr_Lock", 0x0a01},
	{"Control_Lock", 0x0a02},
	{"Alt_Lock", 0x0a03},
	{"ShiftL_Lock", 0x0a04},
	{"ShiftR_Lock", 0x0a05},
	{"CtrlL_Lock", 0x0a06},
	{"CtrlR_Lock", 0x0a07},
	{"SShift", 0x0c00},
	{"SAltGr", 0x0c01},
	{"SControl", 0x0c02},
	{"SAlt", 0x0c03},
	{"SShiftL", 0x0c04},
	{"SShiftR", 0x0c05},
	{"SCtrlL", 0x0c06},
	{"SCtrlR", 0x0c07},
	
	{"Home", 0x0114},
	{"End", 0x0117},
	{"PageUp", 0x0118},
	{"PageDown", 0x0119},
	
	{NULL, 0}
};

/* Attempting to clear keymap before loading new one breaks special keys like
 * shift and control, why?
*/
#if 0
static void clear_keymaps(void) {
	int map, key;
	struct kbentry ke;
	
	ke.kb_value = K_HOLE;
	
	for(map = 0; map < MAX_NR_KEYMAPS; map++) {
		ke.kb_table = map;
		
		for(key = 0; key < NR_KEYS; key++) {
			ke.kb_index = key;
			ioctl(fileno(stdin), KDSKBENT, &ke);
		}
	}
}
#endif

void load_keymap(char const *file) {
	FILE *fh = fopen(file, "r");
	if(!fh) {
		printD("Error opening %s: %s\n", strerror(errno));
		return;
	}
	
	char line[1024], *lp;
	int lnum = 0, i;
	struct kbentry ke;
	
	while(fgets(line, 1024, fh)) {
		lnum++;
		
		lp = line+strspn(line, "\r\n");
		lp[strcspn(lp, "\r\n")] = '\0';
		
		ke.kb_table = atoi(lp);
		lp += strspn(lp, "1234567890");
		lp += strspn(lp, "\t ");
		
		ke.kb_index = atoi(lp);
		lp += strspn(lp, "1234567890");
		lp += strspn(lp, "\t ");
		
		if(*lp) {
			for(i = 0; codes[i].name; i++) {
				if(kl_streq(lp, codes[i].name)) {
					ke.kb_value = codes[i].code;
					break;
				}
			}
			
			if(!codes[i].name) {
				debug("Unknown action at line %d: '%s'", lnum, lp);
				continue;
			}
			
			ioctl(fileno(stdin), KDSKBENT, &ke);
		}
	}
	
	if(ferror(fh)) {
		printD("Error reading %s: %s\n", strerror(errno));
	}
	
	fclose(fh);
}
